@@ -127,12 +127,18 @@ class Engine(object):
             # but use the pickling / unpickling procedure to detect pickling problems
             f_s = dill.dumps(f)
             args_s = dill.dumps(args)
-            result = eval_pickled(f_s, args_s)
+            try:
+                result = eval_pickled(f_s, args_s)
+            except Exception, e:
+                result = DelayedException(e)
             callback(result)
         else:
             run_local = run_local is True or Engine.pool is None
             if run_local:
-                result = f(*args)
+                try:
+                    result = f(*args)
+                except Exception, e:
+                    result = DelayedException(e)
                 callback(result)
             else:
                 f_s = dill.dumps(f)
@@ -141,6 +147,10 @@ class Engine(object):
                                         callback=callback)
 
 
+def _info(listener):
+    return listener.f if hasattr(listener, "f") else listener
+
+
 class DataStream(object):
 
     __metaclass__ = abc.ABCMeta
@@ -150,10 +160,14 @@ class DataStream(object):
         self._queues = dict()
 
     def empty(self, listener):
+        get_logger(self).debug("%s asks for empty input queue", _info(listener))
         return self._queues.get(listener).empty()
 
     def get(self, listener):
-        return self._queues.get(listener).get()
+        item = self._queues.get(listener).get()
+        get_logger(self).debug("%s asks and gets item %d with value %s",
+                               _info(listener), item.number, item.value)
+        return item
 
     @abc.abstractmethod
     def size(self):
@@ -161,7 +175,9 @@ class DataStream(object):
 
     def put(self, item):
         """this is where computation results are passed to."""
-        for queue in self._queues.values():
+        for listener, queue in self._queues.items():
+            get_logger(self).debug("put item %d with value %s to %s",
+                                   item.number, item.value, _info(listener))
             queue.put(item)
 
     def register_listener(self, listener):
@@ -261,11 +277,14 @@ class OutputNode(DataStream):
         """
         self.input_.start_computations()
 
-        results = [self.input_.get(listener=self) for _ in range(self.size())]
-        if not do_not_reraise_exceptions:
-            for result in results:
-                if isinstance(result.value, DelayedException):
-                    result.value.re_raise()
+        results = []
+        for _ in range(self.size()):
+            item = self.input_.get(listener=self)
+            if isinstance(item.value, DelayedException):
+                if not do_not_reraise_exceptions:
+                    item.value.re_raise()
+            results.append(item)
+
         results.sort(key=lambda en_item: en_item.number)
         self.results = [item.value for item in results]
 
@@ -327,61 +346,65 @@ class JoinComputation(Computation):
                 time.sleep(0.001)
 
 
-class SingleInputComputation(Computation):
-
-    def __init__(self, f, inputs, run_local=False, debug=False):
-        assert len(inputs) == 1, "summarize computations only accept one input stream"
-        super(SingleInputComputation, self).__init__(f, inputs, run_local, debug)
-
-
-class SummarizeComputation(SingleInputComputation):
+class SummarizeComputation(Computation):
 
     def size(self):
         return 1
 
-    def _start_computation(self):
+    def _collect_input_data(self):
         items = []
         stream = self.input_streams[0]
         for i in range(stream.size()):
             item = stream.get(listener=self)
             if isinstance(item.value, DelayedException):
                 self.put(item)
-                return
+                return None, None
             items.append(item)
+
+        fixed_args = []
+        for other_stream in self.input_streams[1:]:
+            assert other_stream.size() == 1
+            item = other_stream.get(listener=self)
+            if isinstance(item.value, DelayedException):
+                self.put(item)
+                return None, None
+            fixed_args.append(item)
+
         items.sort(key=lambda item: item.number)
         values = [v for __, v in items]
+        return values, fixed_args
+
+    def _start_computation(self):
+
+        values, fixed_args = self._collect_input_data()
+        if values is None:
+            return
         arg = EnumeratedItem(0, values)
-        self._start_async_computation_for(0, (arg,))
+        args = (arg,) + tuple(fixed_args)
+        self._start_async_computation_for(0, args)
 
 
-class FullStreamComputation(SingleInputComputation):
+class FullStreamComputation(SummarizeComputation):
 
     def size(self):
         return self.input_streams[0].size()
 
     def _start_computation(self):
-        items = []
-        stream = self.input_streams[0]
-        for i in range(stream.size()):
-            item = stream.get(listener=self)
-            if isinstance(item.value, DelayedException):
-                self.put(item)
-                return
-            items.append(item)
-        items.sort(key=lambda item: item.number)
-        values = [v for __, v in items]
-        # arg = EnumeratedItem(0, values)
-        # self._start_async_computation_for(0, (arg,))
-        for i, value in enumerate(values):
-            if isinstance(value, DelayedException):
-                wrap_callback(i, self.put)(value)
-                return
+
+        values, fixed_args = self._collect_input_data()
+
+        if values is None:
+            return
+
+        # this uses the plain engine because we hava a speical callback.
+        # so we have to work with plain data, not EnumeratedItem instances:
 
         def forward_result(result):
             for i, item in enumerate(result):
                 self.put(EnumeratedItem(i, item))
 
-        Engine.run_async(self.f, (values,), forward_result, self.run_local, self.debug)
+        args = (values,) + tuple(a.value for a in fixed_args)
+        Engine.run_async(self.f, args, forward_result, self.run_local, self.debug)
 
 
 class ComputedSource(Computation):
